package org.openxma.dsl.generator.helper;

import java.lang.reflect.Constructor;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.eclipse.emf.mwe.core.resources.ResourceLoaderFactory;
import org.openxma.dsl.common.DslConstants;
import org.openxma.dsl.core.model.DataTypeAndTypeParameter;
import org.openxma.dsl.core.model.Editor;
import org.openxma.dsl.core.model.ParameterDefinition;
import org.openxma.dsl.core.model.ReferenceWithParameter;
import org.openxma.dsl.core.model.SimpleType;
import org.openxma.dsl.core.model.Type;
import org.openxma.dsl.core.model.ValidatorReference;
import org.openxma.dsl.generator.GeneratorConfiguration;
import org.openxma.dsl.pom.model.CustomizeableField;

import at.spardat.xma.guidesign.GuidesignFactory;
import at.spardat.xma.guidesign.PushButton;
import at.spardat.xma.guidesign.XMAComponent;
import at.spardat.xma.guidesign.XMAComposite;
import at.spardat.xma.guidesign.XMAWidget;
import at.spardat.xma.guidesign.flex.FlexFactory;
import at.spardat.xma.guidesign.flex.XMAFunction;

import com.google.common.collect.Sets;

public class SimpleTypeExtension {
	private static Set<String> WRAPPER_TYPES= Sets.newHashSet("Integer","Boolean","Short","Byte","Character","Long","Double","Float");

	/**
	 * Get the DataTypeAndTypeParameter in the definition stack with the name
	 * typeName. In case that there is no definition with this name found the
	 * method returns null.
	 *
	 * @param type
	 *            The data definition
	 * @param typeName
	 *            The name for which a definition is searched
	 * @returnTthe found data definition or null
	 */
	public static DataTypeAndTypeParameter getTypeForName(DataTypeAndTypeParameter type, String typeName) {
		Type dataType = type.getDataType();
		if (dataType instanceof SimpleType) {
			SimpleType simpleType = (SimpleType)dataType;
			if (simpleType.getName().equals(typeName)) {
				return type;
			} else if (simpleType.getTypeDefinition() != null && simpleType.getTypeDefinition().getDefinition() != null) {
				DataTypeAndTypeParameter baseDefinition = simpleType.getTypeDefinition().getDefinition();
				return getTypeForName(baseDefinition, typeName);
			}
			return null;
		} 
		throw new IllegalArgumentException("The type '" + dataType + "' is a not an instance of '"
				+ SimpleType.class + "'");
	}

	/**
	 * Get the most detailed data-Editor in the type definition stack.
	 *
	 * @param type
	 *            The type which for which the editor is searched
	 * @return The data-Editor or null
	 */
	public static Editor getDataEditor(DataTypeAndTypeParameter type) {
		if (type.getDataType() instanceof SimpleType == false) {
			throw new IllegalArgumentException("The type '" + type.getDataType() + "' is a not an instance of '"
					+ SimpleType.class + "'");
		}
		SimpleType simpleType = (SimpleType) type.getDataType();
		if (simpleType.getEditor() != null) {
			return simpleType.getEditor();
		} else if (simpleType.getTypeDefinition() != null && simpleType.getTypeDefinition().getDefinition() != null) {
			DataTypeAndTypeParameter baseDefinition = simpleType.getTypeDefinition().getDefinition();
			return getDataEditor(baseDefinition);
		}
		return null;
	}

	/**
	 * Creates a button to start the custom editor.
	 *
	 * @param f
	 *            The CustomizableField to which the editor belongs
	 * @param dataWidget
	 *            The dataWidget for the CustomizableField f
	 * @return The PushButton to start the custom editor.
	 */
	public static XMAWidget createCustomEditorWidget(CustomizeableField f, String defaultWidgetName, XMAComposite composite) {
		Editor dataEditor = getDataEditor(f.getAttribute().getType());
		if (dataEditor != null) {
			// create button
			PushButton button = GuidesignFactory.eINSTANCE.createPushButton();
			button.setYnGenerated(true);
			button.setNamInstance("btn_" + defaultWidgetName + "_editBy_" + dataEditor.getName());
			button.setLabelName(dataEditor.getLabel());
			String imageUri = GeneratorConfiguration.INSTANCE.getLayoutStrategy().getEditorImageUri(dataEditor);
			if (imageUri != null) {
                button.setUriImage(imageUri);
            }

			// add event-mapping
			XMAComponent xmaComponent = composite.getComponent();
			XMAFunction func = getXMAFunctionByName(xmaComponent, "invokeCustomDataEditor");
			button.getSelectFunction().add(func);
			// return the created button
			return button;
		}
		// no button created because no editor defined
		return null;
	}

	/**
	 * Get the XMA function of a XMAComponent for a defined name
	 *
	 * @param comp
	 * @param functionName
	 * @return
	 */
	public static XMAFunction getXMAFunctionByName(XMAComponent comp, String functionName) {
		// Search for an existing function
		Iterator<XMAFunction> functionsIterator = comp.getFunctions().iterator();
		while (functionsIterator.hasNext()) {
			XMAFunction function = functionsIterator.next();
			if (function.getNamFunction().equals(functionName)) {
				return function;
			}
		}
		// Create the function
		XMAFunction function = FlexFactory.eINSTANCE.createXMAFunction();
		function.setNamFunction(functionName);
		function.setYnServerSide(false);
		comp.getFunctions().add(function);
		return function;
	}

	/**
	 * TODO: replace the calls of this method by direct calls of the according member-function
	 * @param typeAndParameter
	 * @return
	 */
	public static List<ReferenceWithParameter> getDefinitionStack(ReferenceWithParameter typeAndParameter) {
		if (typeAndParameter == null) {
			throw new IllegalArgumentException("The parameter 'typeAndParameter' must not be null");
		}
		return typeAndParameter.getDefinitionStack();
	}	
	
	public static List<ParameterDefinition> getParameterDefinitions(ReferenceWithParameter typeAndParameter) {
		if (typeAndParameter == null) {
			throw new IllegalArgumentException("The parameter 'typeAndParameter' must not be null");
		}
		return typeAndParameter.getParameterDefinitions();
	}	

	/**
	 * Get the ValidatorReference definition and determine the definition-stack.
	 * (Use this method in all cases where the definition-stack in not already
	 * available.)
	 *
	 * @param dataTypeAndTypeParameter
	 * @return
	 */
	public static ValidatorReference getValidatorReferenceDefinition(DataTypeAndTypeParameter dataTypeAndTypeParameter) {
		// List<DataTypeAndTypeParameter> definitionStack = dataTypeAndTypeParameter.getDefinitionStack();
		return dataTypeAndTypeParameter.getValidatorReferenceDefinition();
	}
	

	/**
	 * @param dataTypeAndTypeParameter
	 * @return the root ValidatorReference definition
	 */
	public static ValidatorReference getRootValidatorReference(DataTypeAndTypeParameter dataTypeAndTypeParameter) {
		return dataTypeAndTypeParameter.getValidatorReferenceDefinition().getRootValidatorReference();
	}


	public static String getJavaType(final SimpleType simpleType) {
		return simpleType.getInstanceTypeName();
	}
	
	public static Boolean isPrimitiveWrapperType(final SimpleType simpleType) {
		String typeName = simpleType.getInstanceTypeName();
		return null!=typeName ? WRAPPER_TYPES.contains(typeName) : false;
	}
	
	public static Boolean isPrimitiveWrapperType(String typeName) {
		return null!=typeName ? WRAPPER_TYPES.contains(typeName) : false;
	}

	public static String getNativeJavaTypeForValueObject(String valueObject) {
		// example: org.openxma.dsl.platform.valueobject.ValueObjectString
		try {

			Class valueObjectClass = ResourceLoaderFactory.createResourceLoader().loadClass(valueObject);

			if (valueObjectClass == null) {
				valueObjectClass = Thread.currentThread().getContextClassLoader().loadClass(valueObject);
			}
				
	    	Constructor[] constructors = valueObjectClass.getConstructors();
	    	for (int i = 0; i < constructors.length; i++) {
				if(constructors[i].getParameterTypes().length == 1) {
					if(byte[].class.equals(constructors[i].getParameterTypes()[0])) {
						return "byte[]";
					}
					return substringJavaLangPackage(constructors[i].getParameterTypes()[0].getName());
				}
			}
	    	throw new IllegalStateException("No correct Constructor found in "+valueObject);
		} catch (ClassNotFoundException e) {
			throw new IllegalStateException("ValueObject-Class '"
					+ valueObject + "' not found");
		}
	}

	public static String generateVariableNameForValueObject(String valueObject) {
		String type = getNativeJavaTypeForValueObject(valueObject);
		String[] split = type.split("\\.");
		type = split.length == 0 ? type : split[split.length-1];
		type = "p"+type.substring(0,1).toUpperCase()+type.substring(1);
		return type.replaceAll("\\[\\]", "Array");
	}

	private static String substringJavaLangPackage(String s) {
		if(s.startsWith(DslConstants.PACKAGE_JAVA_LANG)) {
			return s.substring(DslConstants.PACKAGE_JAVA_LANG.length());
		}
		return s.replaceAll("\\^", "");
	}
	

}

